/* GnoBang: (c) 2008 by Andreas S. Krebs
 * 
 * This file is part of GnoBang.
 *
 * GnoBang is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * GnoBang is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "gnobang.h"

#include "pixmaps/chip.xpm"

/* the lightness distance for the shadow colors in HLS color space */
#define SHADOWCOLOR_LIGHTNESS_DISTANCE	0.12f

/* local prototypes */

static GdkImage	*getImageFromXPM(GdkWindow *win, char **data);
static void showWinMessage(gboolean won, int playerNr);
static void getUndoImages();
static void drawShadowRectangle(int x1, int y1, int x2, int y2, guint32 topColor, guint32 bottomColor);
static void RGBtoHLS(int r, int g, int b, float *ph, float *pl, float *ps);
static int HLStoValue(float min, float max, float mmm, float h);
static void HLStoRGB(float h, float l, float s, int *pr, int *pg, int *pb);


/* local variables */

static GdkPixmap	*gfxPixmap=NULL;
static GdkGC    	*image_pixmapGC=NULL;
static GdkGC		*draw_pixmapGC=NULL;
static GdkGC		*win_pixmapGC=NULL;
static GdkGC		*image_windowGC=NULL;
static GdkImage 	*playerImage[3]={ NULL, NULL, NULL };
static GdkImage		*undoImage[3][3]={ { NULL, NULL, NULL }, { NULL, NULL, NULL }, { NULL, NULL, NULL } };

static GdkWindow	*gfxWindow=NULL;
static gboolean		gotImages=FALSE;


/************************************************************************
 *																		*
 * create images from xpm data										 	*
 *  																	*
 ************************************************************************/

static char	cols_txt[][30]={ "  c #", ". c #", "X c #", "o c #" };

void createXImages()
{
	guint32	topCol, col, botCol;
	
	chip_xpm[1]=cols_txt[0];
	chip_xpm[2]=cols_txt[1];
	chip_xpm[3]=cols_txt[2];
	chip_xpm[4]=cols_txt[3];

	setcolor(&cols_txt[0][5], goGame->BoardBackground);
	
	getShadowColors(goGame->Player[0].Color, &topCol, &col, &botCol);
	setcolor(&cols_txt[1][5], topCol);
	setcolor(&cols_txt[2][5], col);
	setcolor(&cols_txt[3][5], botCol); 
	playerImage[0]=getImageFromXPM(gnoBang->drawingArea->window, chip_xpm);

	getShadowColors(goGame->Player[1].Color, &topCol, &col, &botCol);
	setcolor(&cols_txt[1][5], topCol);
	setcolor(&cols_txt[2][5], col);
	setcolor(&cols_txt[3][5], botCol); 
	playerImage[1]=getImageFromXPM(gnoBang->drawingArea->window, chip_xpm);

	getShadowColors(goGame->HintColor, &topCol, &col, &botCol);
	setcolor(&cols_txt[1][5], topCol);
	setcolor(&cols_txt[2][5], col);
	setcolor(&cols_txt[3][5], botCol); 
	playerImage[2]=getImageFromXPM(gnoBang->drawingArea->window, chip_xpm);
}


/************************************************************************
 *																		*
 * set rgb-text for according to the color defined by <col>	 			*
 *  																	*
 ************************************************************************/

void setcolor(char *cs, guint32 col)
{
	int    r, g, b;

	r=(col>>16)&0xFF;
	g=(col>>8)&0xFF;
	b=(col)&0xFF;

	sprintf(cs, "%x%x%x%x%x%x", (r/0x10), (r%0x10), (g/0x10), (g%0x10), (b/0x10), (b%0x10) );
}


/************************************************************************
 *																		*
 * get GdkColor for given RGB value									 	*
 *  																	*
 ************************************************************************/

void setGdkColor(GtkWidget *widget, GdkColor *gdkCol, guint32 col)
{
	GdkColormap	*cmap=gtk_widget_get_colormap(widget);
	
	gdkCol->red  =(col&0xff0000)>>8;
	gdkCol->green=(col&0x00ff00);
	gdkCol->blue =(col&0x0000ff)<<8;
	
	gdk_rgb_find_color(cmap, gdkCol);
}


/************************************************************************
 *																		*
 * get RGB value from given GdkColor								 	*
 *  																	*
 ************************************************************************/

guint32 getGdkColor(GdkColor *gdkCol)
{
	guint32	r, g, b;
	
	r=(gdkCol->red  &0x00ff00)<<8;
	g=(gdkCol->green&0x00ff00);
	b=(gdkCol->blue &0x00ff00)>>8;
	
	return (r|g|b);
}


/************************************************************************
 *																		*
 * get the RGB values for drawing shadows based on the given base color	*
 *  																	*
 ************************************************************************/

void getShadowColors(guint32 baseCol, guint32 *topCol, guint32 *col, guint32 *botCol)
{
	int		r, g, b;
	float	h, l, s;
	
	r=(baseCol&0xff0000)>>16;
	g=(baseCol&0x00ff00)>>8;
	b=(baseCol&0x0000ff);

	RGBtoHLS(r, g, b, &h, &l, &s);
	
	// limit l, so that we can create shadow colors
	if(l<SHADOWCOLOR_LIGHTNESS_DISTANCE)
		l=SHADOWCOLOR_LIGHTNESS_DISTANCE;
	if(l>(1.0f-SHADOWCOLOR_LIGHTNESS_DISTANCE))
		l=1.0f-SHADOWCOLOR_LIGHTNESS_DISTANCE;
	
	// normal color
	HLStoRGB(h, l, s, &r, &g, &b);
	*col=(r<<16)|(g<<8)|b;
	
	// top (lighter color)
	HLStoRGB(h, l+SHADOWCOLOR_LIGHTNESS_DISTANCE, s, &r, &g, &b);
	*topCol=(r<<16)|(g<<8)|b;

	// bottom (darker color)
	HLStoRGB(h, l-SHADOWCOLOR_LIGHTNESS_DISTANCE, s, &r, &g, &b);
	*botCol=(r<<16)|(g<<8)|b;
}


/************************************************************************
 *																		*
 * create a GdkImage from the given XPM data						 	*
 *  																	*
 ************************************************************************/

static GdkImage	*getImageFromXPM(GdkWindow *win, char **data)
{
	GdkPixmap	*pixmap;
	GdkImage	*image;
	gint		w, h;

	pixmap=gdk_pixmap_create_from_xpm_d(win, NULL, NULL, data);

	if(pixmap==NULL)
	{
		showErrorMessage("Failed to create image from XPM.");
		return NULL;
	}
	
	gdk_window_get_size(pixmap, &w, &h);

	image=gdk_image_get(pixmap, 0, 0, w, h);

	gdk_pixmap_unref(pixmap);

	return image;	
}


/************************************************************************
 *																		*
 * create and initialize all needed GCs								 	*
 *  																	*
 ************************************************************************/

void initGCs()
{
	gint	depth;
	
	gfxWindow=gnoBang->drawingArea->window;
	depth=gdk_drawable_get_depth(gfxWindow);

	/* free old GC if allocated before */
	if(win_pixmapGC)
		gdk_gc_unref(win_pixmapGC);
	if(draw_pixmapGC)
		gdk_gc_unref(draw_pixmapGC);
	if(image_windowGC)
		gdk_gc_unref(image_windowGC);
	if(image_pixmapGC)
		gdk_gc_unref(image_pixmapGC);
	
	if(gfxPixmap)
		gdk_pixmap_unref(gfxPixmap);
	
	/* now alloc the GCs again */
	gfxPixmap=gdk_pixmap_new(gnoBang->drawingArea->window, goGame->Width, goGame->Height, depth);
	
	image_pixmapGC=gdk_gc_new(gfxPixmap);	
	gdk_rgb_gc_set_foreground(image_pixmapGC, goGame->BoardColor);
	gdk_rgb_gc_set_background(image_pixmapGC, goGame->BoardBackground);
	
	image_windowGC=gdk_gc_new(gfxWindow);	
	gdk_rgb_gc_set_foreground(image_windowGC, goGame->BoardColor);
	gdk_rgb_gc_set_background(image_windowGC, goGame->BoardBackground);

	draw_pixmapGC=gdk_gc_new(gfxPixmap);	
	gdk_rgb_gc_set_foreground(draw_pixmapGC, goGame->BoardBackground);

	win_pixmapGC=gdk_gc_new(gfxPixmap);	
	gdk_rgb_gc_set_foreground(win_pixmapGC, goGame->BoardBackground);
	gdk_gc_set_line_attributes(win_pixmapGC, IMAGE_SIZE-8, GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_ROUND);
}


/************************************************************************
 *																		*
 * draw a chip for player <playerNo> at the given field <x>, <y>		*
 *  																	*
 ************************************************************************/

void drawChipAt(int x, int y, int playerNo)
{
	static int hx=-1, hy=-1;	/* coordinates for hint */
  
	if(hx>=0 && hy>=0)
	{
		undoChipAt(hx, hy);
    
		hx=hy=-1;
	}

	if(playerNo==PLAYER_HINT)
	{
		hx=x;
		hy=y;
	}
	else
		setField(x, y, playerNo+1);
  
	x*=FIELD_SIZE;
	x+=BORDER_SIZE+1;
  
	y*=FIELD_SIZE;
	y+=BORDER_SIZE+1;
 
	gdk_draw_image(gfxPixmap, image_pixmapGC, playerImage[playerNo], 0, 0, x, y, IMAGE_SIZE, IMAGE_SIZE);
	gdk_draw_image(gfxWindow, image_windowGC, playerImage[playerNo], 0, 0, x, y, IMAGE_SIZE, IMAGE_SIZE);
}


/************************************************************************
 *																		*
 * remove a chip from the given field <x>, <y>						 	*
 *  																	*
 ************************************************************************/

void undoChipAt(int x, int y)
{
	int i, j;

	setField(x, y, FIELD_EMPTY);
  
	if(x==0) 
		i=0;
	else if(x==(goGame->Columns-1))
		i=2;
	else
		i=1;

	if(y==0) 
		j=0;
	else if(y==(goGame->Rows-1))
		j=2;
	else
		j=1;
  
	x*=FIELD_SIZE;
	x+=BORDER_SIZE+1;
  
	y*=FIELD_SIZE;
	y+=BORDER_SIZE+1;
  
	gdk_draw_image(gfxPixmap, image_pixmapGC, undoImage[i][j], 0, 0, x, y, IMAGE_SIZE, IMAGE_SIZE);
	gdk_draw_image(gfxWindow, image_windowGC, undoImage[i][j], 0, 0, x, y, IMAGE_SIZE, IMAGE_SIZE);
}


/************************************************************************
 *																		*
 * draw the game board 												 	*
 *  																	*
 ************************************************************************/

void drawBoard()
{
	GdkRectangle	clipRec;
	guint32			topCol, col, botCol;
	int				i, j, x, y;

	getShadowColors(goGame->BoardColor, &topCol, &col, &botCol);
	
	clipRec.x=0;
	clipRec.y=0;
	clipRec.width=goGame->Width;
	clipRec.height=goGame->Height;

    gdk_gc_set_clip_rectangle(draw_pixmapGC, &clipRec);

    // first clear the background
	gdk_rgb_gc_set_foreground(draw_pixmapGC, goGame->BoardBackground);
	gdk_draw_rectangle(gfxPixmap, draw_pixmapGC, TRUE, 0, 0, goGame->Width, goGame->Height);
  
	gdk_rgb_gc_set_foreground(draw_pixmapGC, col);

	gdk_draw_rectangle(gfxPixmap, draw_pixmapGC, FALSE, FIELD_OFFSET-1, FIELD_OFFSET-1, 
		 (goGame->Columns-1)*FIELD_SIZE+2, (goGame->Rows-1)*FIELD_SIZE+2);

	drawShadowRectangle(FIELD_OFFSET-2, FIELD_OFFSET-2, 
	   	      goGame->Width-FIELD_OFFSET+1, goGame->Height-FIELD_OFFSET+1,		      
		      topCol, botCol);

	gdk_rgb_gc_set_foreground(draw_pixmapGC, col);
	x=y=FIELD_OFFSET;
	for(i=0;i<goGame->Columns;i++)
	{
		gdk_draw_line(gfxPixmap, draw_pixmapGC, x, FIELD_OFFSET, x, goGame->Height-FIELD_OFFSET);
		x+=FIELD_SIZE;
	}

	for(i=0;i<goGame->Rows;i++)
	{
		gdk_draw_line(gfxPixmap, draw_pixmapGC, FIELD_OFFSET, y, goGame->Width-FIELD_OFFSET, y);
		y+=FIELD_SIZE;
	}

	x=FIELD_OFFSET+1;
	for(i=0;i<(goGame->Columns-1);i++)
	{
		y=FIELD_OFFSET+1;
		for(j=0;j<(goGame->Rows-1);j++)
		{
			drawShadowRectangle(x, y, x+FIELD_SIZE-2, y+FIELD_SIZE-2, botCol, topCol);
			y+=FIELD_SIZE;
		}

		x+=FIELD_SIZE;
	}
	
	getUndoImages();
	
	gdk_draw_pixmap(gfxWindow, image_windowGC, gfxPixmap, 0, 0,
			0, 0, goGame->Width, goGame->Height);
}



/************************************************************************
 *																		*
 * get the 9 possible images, to remove a chip from a field				*
 *  																	*
 ************************************************************************/

static void getUndoImages()
{
	int		i, j, x, y;

	if(!gotImages)
	{
		for(i=0; i<3; i++)
		{
			x=(i>=2) ? (goGame->Columns-1) : i;

			x*=FIELD_SIZE;
			x+=BORDER_SIZE+1;

			for(j=0; j<3; j++)
			{
				y=(j>=2) ? (goGame->Rows-1) : j;

				y*=FIELD_SIZE;
				y+=BORDER_SIZE+1;
				
				undoImage[i][j]=gdk_image_get(gfxPixmap, x, y, IMAGE_SIZE, IMAGE_SIZE);				
			}
		}
		
		gotImages=TRUE;
	}
}


/************************************************************************
 *																		*
 * draw a rectangle with a top and bottom shadow				 	 	*
 *  																	*
 ************************************************************************/

static void drawShadowRectangle(int x1, int y1, int x2, int y2, 
								guint32 topColor, guint32 bottomColor)
{
	int h;
  
	if(x1 > x2) { h=x1; x1=x2; x2=h; }
	if(y1 > y2) { h=y1; y1=y2; y2=h; }

	gdk_rgb_gc_set_foreground(draw_pixmapGC, topColor);

	gdk_draw_line(gfxPixmap, draw_pixmapGC, x1, y2, x1, y1);
	gdk_draw_line(gfxPixmap, draw_pixmapGC, x1, y1, x2, y1);

	gdk_rgb_gc_set_foreground(draw_pixmapGC, bottomColor);

	gdk_draw_line(gfxPixmap, draw_pixmapGC, x2, y1+1, x2, y2);
	gdk_draw_line(gfxPixmap, draw_pixmapGC, x2, y2, x1+1, y2);  
}


/************************************************************************
 *																		*
 * display Game Over and winner if one exists						 	*
 *  																	*
 ************************************************************************/

void displayGameOver(gboolean won, int playerNr, int x1, int y1, int x2, int y2)
{
	int x_1, y_1, x_2, y_2;
  
	if(won)
	{
		int	opn=OTHER_PLAYER(playerNr);
		
		x_1=x1*FIELD_SIZE +BORDER_SIZE+1+IMAGE_SIZE/2;
		y_1=y1*FIELD_SIZE +BORDER_SIZE+1+IMAGE_SIZE/2; 
		x_2=x2*FIELD_SIZE +BORDER_SIZE+1+IMAGE_SIZE/2;
		y_2=y2*FIELD_SIZE +BORDER_SIZE+1+IMAGE_SIZE/2;
	  
		gdk_rgb_gc_set_foreground(win_pixmapGC, goGame->Player[playerNr].Color);
		gdk_draw_line(gfxPixmap, win_pixmapGC, x_1, y_1, x_2, y_2);
	   
		gdk_draw_pixmap(gfxWindow, image_windowGC, gfxPixmap, 0, 0,
						0, 0, goGame->Width, goGame->Height);
		
	    if(goGame->Player[playerNr].Type==PLAYER_COMPUTER && goGame->Player[opn].Type==PLAYER_HUMAN)
	    	playSound(LOSE_SOUND, FALSE);
	    else if(goGame->Player[playerNr].Type==PLAYER_HUMAN && goGame->Player[opn].Type==PLAYER_COMPUTER)
	    	playSound(WIN_SOUND, FALSE);      
	    else
	    	playSound(OVER_SOUND, FALSE);
	}
	else
	{
	    playSound(OVER_SOUND, FALSE);
	}
	
    showWinMessage(won, playerNr);    
}


/************************************************************************
 *																		*
 * show a winner message											 	*
 *  																	*
 ************************************************************************/

static void showWinMessage(gboolean won, int playerNr)
{
	GtkWidget	*dialog;
	char		*msg=won ? "<span size=\"large\">The winner is: </span><span size=\"large\" color=\"#%s\">%s</span>" : "<span size=\"large\">The game ended draw !</span>";
	
	dialog=gtk_message_dialog_new_with_markup(gnoBang->window,
            GTK_DIALOG_DESTROY_WITH_PARENT,
            GTK_MESSAGE_INFO,
            GTK_BUTTONS_OK,
            msg, gnoBang->playerColor[playerNr], goGame->Player[playerNr].Name);
	
	gtk_window_set_title(GTK_WINDOW(dialog), "Game Over");
	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
}


/************************************************************************
 *																		*
 * free all pixmaps and images										 	*
 *  																	*
 ************************************************************************/

void destroyAll()
{
	int	i,j;
  
	if(win_pixmapGC)
		gdk_gc_unref(win_pixmapGC);
	if(image_pixmapGC)
		gdk_gc_unref(image_pixmapGC);
	if(draw_pixmapGC)
		gdk_gc_unref(draw_pixmapGC);
	if(image_windowGC)
		gdk_gc_unref(image_windowGC);
	win_pixmapGC=image_pixmapGC=draw_pixmapGC=image_windowGC=NULL;
	
	if(gfxPixmap)
		gdk_pixmap_unref(gfxPixmap);
	gfxPixmap=NULL;
	
	for(i=0;i<3;i++)
	{
		if(playerImage[i])
			gdk_image_unref(playerImage[i]);
		
		playerImage[i]=NULL;
	}

	for(i=0;i<3;i++)
	{
		for(j=0;j<3;j++)
		{
			if(undoImage[i][j])
				gdk_image_unref(undoImage[i][j]);
			
			undoImage[i][j]=NULL;
		}
	}
	
	gotImages=FALSE;
}


/************************************************************************
 *																		*
 * expose callback for the drawing area 								*
 *   																	*
 ************************************************************************/

gboolean exposeCB(GtkWidget *map, GdkEventExpose *event)
{
	gdk_draw_pixmap(gfxWindow, image_windowGC, gfxPixmap, event->area.x, event->area.y,
			event->area.x, event->area.y,event->area.width, event->area.height);
	
	return TRUE;
}


/************************************************************************
 *																		*
 * input callback procedure for drawing area:						 	*
 *   this procedure recognizes moves made by human players 				*
 ************************************************************************/

gboolean gfxinputCB(GtkWidget *widget, GdkEventButton *event) 
{
	if(event->type==GDK_BUTTON_RELEASE && goGame->GameStatus!=GAME_FINISHED &&
			goGame->Player[goGame->PlayersTurn].Type==PLAYER_HUMAN)
	{
		if(event->button==1)
		{
			int	x, y;

			x=(int)event->x -BORDER_SIZE;
			y=(int)event->y -BORDER_SIZE;
    
			x/=FIELD_SIZE;
			y/=FIELD_SIZE;
			
			if(x>=0 && x<goGame->Columns && y>=0 && y<goGame->Rows)
			{
				if(getField(x, y)==FIELD_EMPTY)
					makeMove(x, y);
				else
					playSound(ILLEGAL_SOUND, FALSE);
			}
		}
	}
	
	return TRUE;
}


/************************************************************************
 *																		*
 * the methods for converting between HLS and RGB		 				*
 * 																		*
 ************************************************************************/

static void RGBtoHLS(int r, int g, int b, float *ph, float *pl, float *ps)
{	
	int		min, max;
	float	h, l, s, sum, dif;
	
	max=MAX(r, MAX(g, b));
	min=MIN(r, MIN(g, b));
	
	sum=(float)(max+min);
	dif=(float)(max-min);

	l=sum/510.0f;
	
	if(max==min)
	{
		s=h=0.0f;
	}
	else
	{
		if(l<=0.5)
			s=dif/sum;
		else
			s=dif/(510.0f-sum);

		if(r==max)
			h=6.0f+(g-b)/dif;
		else if(g==max)
			h=2.0f+(b-r)/dif;
		else
			h=4.0f+(r-g)/dif;
		h*=60.0f;
		
		if(h>360.0f)
			h=h-360.0f;
	}

	(*ph)=h;
	(*pl)=l;
	(*ps)=s;
}

static int HLStoValue(float min, float max, float mmm, float h)
{
	float	val;
	
	if(h>360.0f)
		h-=360.0f;
	else if(h<0.0f)
		h+=360.0f;
	
	if(h<60.0f)
		val=min+mmm*h;
	else if(h<180.0f)
		val=max;
	else if(h<240.0f)
		val=min+mmm*(240.0f-h);
	else
		val=min;
	
	val*=255.0f;
	
	return (int)val;
}

static void HLStoRGB(float h, float l, float s, int *pr, int *pg, int *pb)
{
	int	r, g, b;

	if(s==0.0f)
	{
		r=g=b=(int)(l*255.0f);
	}
	else
	{
		float min, max, mmm;
			
		if(l<=0.5f)
			max=l+l*s;
		else
			max=l+s-l*s;
		
		min=2.0f*l-max;
		mmm=(max-min)/60.0f;
		
		r=HLStoValue(min, max, mmm, h+120.0f);
		g=HLStoValue(min, max, mmm, h);
		b=HLStoValue(min, max, mmm, h-120.0f);
	}
	
	(*pr)=r;
	(*pg)=g;
	(*pb)=b;
}

