/* GnoBang: (c) 2008 by Andreas S. Krebs
 * 
 * This file is part of GnoBang.
 *
 * GnoBang is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * GnoBang is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GnoBang.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "gnobang.h"

#define COMPUTER_MOVETIME	20	/* leave a little time to system */

/* some local prototypes */
static void initBoard();
static gboolean computerOnlyGameTimerCB(gpointer data);

static void makeComputerMove();
static void calculateNextMove(int *, int *, int);
static gboolean checkFor5(int, int, int *, int *, int *, int *);

static long getFieldValue(int, int, int);
static long getAreaValue(DirInfo *);
static void getDirectionInfo(int, int, int, int, DirInfo *); 


/* some local variables */
static GoBoard XGoBoard=
{
	NULL,		// BoardField
	NULL,		// MovesDone
	0,			// MoveNumber
	0,			// LastMoveDone
	0			// NumberOfFields
};

/************************************************************************
 *																		*
 * Callback procedure for the `undo`-button:							*
 *   this procedure undoes the last done move							*
 *																		*
 ************************************************************************/

void undoCB(GtkObject *object)
{
	if(XGoBoard.MoveNumber>0)
	{
		XGoBoard.MoveNumber--;    
		undoChipAt(XGoBoard.MovesDone[XGoBoard.MoveNumber*2],
				   XGoBoard.MovesDone[XGoBoard.MoveNumber*2+1]);

		if(goGame->GameStatus!=GAME_PAUSED)
			setGameStatus(GAME_PAUSED);
    
		setPlayersTurn(OTHER_PLAYER(goGame->PlayersTurn));     

		if(XGoBoard.MoveNumber==0)
			switchOffWidget(UNDO_WIDGET);
		if(XGoBoard.MoveNumber+1 == XGoBoard.LastMoveDone)
			switchOnWidget(REDO_WIDGET);
	}
}


/************************************************************************
 *																		*
 * Callback procedure for the `redo`-button:   				 			*
 *   this procedure redoes the last undone move 						*
 *																		*
 ************************************************************************/

void redoCB(GtkObject *object)
{
	if(XGoBoard.LastMoveDone > XGoBoard.MoveNumber)
	{
		drawChipAt(XGoBoard.MovesDone[XGoBoard.MoveNumber*2],
				   XGoBoard.MovesDone[XGoBoard.MoveNumber*2+1], goGame->PlayersTurn);
    
		XGoBoard.MoveNumber++;    
		setPlayersTurn(OTHER_PLAYER(goGame->PlayersTurn));
    
		if(XGoBoard.MoveNumber >= XGoBoard.LastMoveDone)
			switchOffWidget(REDO_WIDGET);
		if(XGoBoard.MoveNumber == 1)
			switchOnWidget(UNDO_WIDGET);
	}
}


/************************************************************************
 *																		*
 * Callback procedure for the `hint`-button:							*
 *   this procedure gives a human player a hint for the next move		*
 *																		*
 ************************************************************************/

void hintCB(GtkObject *object)
{
	if(goGame->GameStatus!=GAME_PAUSED)
		setGameStatus(GAME_PAUSED);
    
	if(XGoBoard.MoveNumber < XGoBoard.NumberOfFields)
	{
		int x,y;
    
		calculateNextMove(&x, &y, goGame->PlayersTurn);
		drawChipAt(x, y, PLAYER_HINT);
	}
	else     
		 switchOffWidget(HINT_WIDGET);
}


/************************************************************************
 *																		*
 * Callback procedure for the `control`-button: 						*
 *   this procedure changes the GameStatus depending on the current		*
 *   GameStatus															*
 *																		*
 ************************************************************************/

void controlCB(GtkObject *object)
{
	switch(goGame->GameStatus)
	{
		case GAME_PAUSED:
			if(XGoBoard.MoveNumber != XGoBoard.LastMoveDone)
			{
				switchOffWidget(REDO_WIDGET);
				XGoBoard.LastMoveDone=XGoBoard.MoveNumber;
			}
      
		case GAME_NEW:
			setGameStatus(GAME_RUNNING);
			startPlayerTimer(goGame->PlayersTurn);

			if(goGame->Player[0].Type==PLAYER_COMPUTER && goGame->Player[1].Type==PLAYER_COMPUTER)
			{	
				gdk_threads_add_timeout(COMPUTER_MOVETIME, computerOnlyGameTimerCB, NULL);
			}
			else if(goGame->Player[goGame->PlayersTurn].Type==PLAYER_COMPUTER)
				makeComputerMove(goGame->PlayersTurn);
			break;
			
		case GAME_RUNNING:
			setGameStatus(GAME_PAUSED);
			break;
			
		case GAME_FINISHED:
			newGame();
			break;
	}
}


/************************************************************************
 *																		*
 * initialize a new game												*
 *																		*
 ************************************************************************/

void newGame()
{
	int	playerNr;
	
	setGameStatus(GAME_NEW);
	initBoard();

	switchOffWidget(UNDO_WIDGET);
	switchOffWidget(REDO_WIDGET);
	switchOffWidget(HINT_WIDGET);    

	drawBoard();

	goGame->Player[0].Seconds=0;
	setPlayerTimer(0);
	goGame->Player[1].Seconds=0;
	setPlayerTimer(1);

	playerNr=goGame->StartPlayer;
	if(playerNr==START_WINNER)
		playerNr=goGame->LastWinner;
	else if(playerNr==START_LOOSER)
		playerNr=OTHER_PLAYER(goGame->LastWinner);
	
	setPlayersTurn(playerNr);
}


/************************************************************************
 *																		*
 * alloc memory for a the game board and the moves stack				*
 *																		*
 ************************************************************************/

void allocBoard()
{
	freeBoard();
	
	XGoBoard.BoardField=(char *)malloc(sizeof(char)*goGame->Columns*goGame->Rows);  
	XGoBoard.MovesDone=(char *)malloc(sizeof(char)*goGame->Columns*goGame->Rows*2);   
	
	initBoard();
}


/************************************************************************
 *																		*
 * free memory for a the game board and the moves stack					*
 *																		*
 ************************************************************************/

void freeBoard()
{
	if(XGoBoard.BoardField)
		free(XGoBoard.BoardField);
	XGoBoard.BoardField=NULL;
	
	if(XGoBoard.MovesDone)
		free(XGoBoard.MovesDone);
	XGoBoard.BoardField=NULL;
}


/************************************************************************
 *																		*
 * initialize board by clearing all fields and resetting moves stack	*
 *																		*
 ************************************************************************/

static void initBoard()
{
	/* at beg. all field are empty */
	memset(XGoBoard.BoardField, FIELD_EMPTY, sizeof(char)*goGame->Rows*goGame->Columns);		
	
	XGoBoard.MoveNumber=0;
	XGoBoard.LastMoveDone=0;
	XGoBoard.NumberOfFields=goGame->Rows*goGame->Columns;
}


/************************************************************************
 *																		*
 * start a timer to make computer moves									*
 *																		*
 ************************************************************************/

static gboolean computerOnlyGameTimerCB(gpointer data)
{
	if(goGame->GameStatus==GAME_RUNNING)
	{
		makeComputerMove();
    
		return TRUE;		// call again	    
	}
	else
	{
		return FALSE;		// destroy timer
	}
}


/************************************************************************
 *																		*
 * place a "button" on the specified field						 		*
 *																		*
 ************************************************************************/

void setField(int x, int y, int value)
{
	XGoBoard.BoardField[y*goGame->Columns + x]=(char)value;
}


/************************************************************************
 *																		*
 * return type of the "button" on the specified field				 	*
 *																		*
 ************************************************************************/

int getField(int x, int y)
{
	if(x<0 || x>=goGame->Columns || y<0 || y>=goGame->Rows)
		return FIELD_BORDER;
	else
	{
		return (int)XGoBoard.BoardField[y*goGame->Columns + x];
	}
}


/************************************************************************
 *																		*
 * make a move and change the PlayersTurn 							 	*
 *																		*
 ************************************************************************/

void makeMove(int x, int y)
{
	gboolean	won;
	int			s_x=0, s_y=0, d_x=0, d_y=0;
	int			pno=goGame->PlayersTurn;
	int			opno=OTHER_PLAYER(goGame->PlayersTurn);

	if(x<0 || x>=goGame->Columns || y<0 || y>=goGame->Rows)
	{
		char	*msg;
		
		setGameStatus(GAME_PAUSED);
		
		msg=g_strdup_printf("INTERNAL ERROR: illegal move at %i,%i from player %i", x, y, pno);
		showErrorMessage(msg);
		g_free(msg);
		return;
	}
	
	XGoBoard.MovesDone[XGoBoard.MoveNumber*2]=(unsigned char)x;
	XGoBoard.MovesDone[XGoBoard.MoveNumber*2+1]=(unsigned char)y;
  
	drawChipAt(x, y, pno);
  
	if(XGoBoard.MoveNumber==0)
		switchOnWidget(UNDO_WIDGET);
  
	XGoBoard.MoveNumber++;
	XGoBoard.LastMoveDone=XGoBoard.MoveNumber;

	if(goGame->GameStatus!=GAME_RUNNING)
		setGameStatus(GAME_RUNNING);

	if(goGame->Player[opno].Type==PLAYER_COMPUTER && 
	   goGame->Player[pno].Type==PLAYER_COMPUTER)
	{
		goGame->PlayersTurn=opno;
		/* don't display new turn, cause computer only game is to fast */
	}
	else
	{
		setPlayersTurn(opno); 
	}
	
	/* the player has changed */
	pno=goGame->PlayersTurn;
	opno=OTHER_PLAYER(goGame->PlayersTurn);
	
	won=checkFor5(x, y, &s_x, &s_y, &d_x, &d_y);
	if(won==TRUE || XGoBoard.MoveNumber >= XGoBoard.NumberOfFields)
	{
		setGameStatus(GAME_FINISHED);
		switchOffWidget(UNDO_WIDGET);
	    setCursor(GAME_FINISHED);		/* no more moves possible */

	    goGame->LastWinner=opno;
		displayGameOver(won, opno, s_x, s_y, d_x, d_y);
	    return;
	}
    
	if(goGame->Player[opno].Type==PLAYER_HUMAN && 
       goGame->Player[pno].Type==PLAYER_COMPUTER)
	{
		makeComputerMove(pno);
	}
}


/************************************************************************
 *																		*
 * let computer make a move  										 	*
 *																		*
 ************************************************************************/

static void makeComputerMove()
{
	int x, y;
	
	if(XGoBoard.MoveNumber < XGoBoard.NumberOfFields)
	{
	    /* computer has a one second handicap per move ;-) */
		goGame->Player[goGame->PlayersTurn].Seconds++;
		setPlayerTimer(goGame->PlayersTurn);
	    
	    calculateNextMove(&x, &y, goGame->PlayersTurn); 
	    makeMove(x, y);
	}
	else
	{
		setGameStatus(GAME_FINISHED);
		switchOffWidget(UNDO_WIDGET);
	}
}


/************************************************************************
 *																		*
 * calculate the next move for player playerNr 						 	*
 *																		*
 ************************************************************************/

static void calculateNextMove(int *x, int *y, int playerNr)
{
	int		i, j;
	long	value=-1, tmpval;
	int		rand_num=1;
  
	/* set inital values to something illegal */
	(*x)=-1;
	(*y)=-1;
	
	srand(time(NULL));
  
	for(i=0;i<goGame->Columns;i++)
	{
		for(j=0;j<goGame->Rows;j++)
		{
			if(getField(i,j)==FIELD_EMPTY)
			{
				tmpval=getFieldValue(i, j, playerNr);
	
				if(tmpval > value)
				{
					value=tmpval;
					*x=i;
					*y=j;
	  
					rand_num=1;
				}
				else if(tmpval==value)
				{
					rand_num++;
					if(rand() < RAND_MAX/rand_num)
					{
						*x=i;
						*y=j;
					}
				}
			}
		}      
	}
}


/************************************************************************
 *																		*
 * get value of this field, the field with the highest value is the 	*
 * best move to do														*
 *																		*
 ************************************************************************/

static long getFieldValue(int x, int y, int playerNr)
{
	DirInfo	me_dir, other_dir;
	long	value=0, v;
	int		k;
 
	for(k=0;k<4;k++)
	{
		getDirectionInfo(k, x, y, playerNr, &me_dir);
		getDirectionInfo(k, x, y, OTHER_PLAYER(playerNr), &other_dir);  
  
		if(me_dir.SameColor>=4)
		{  
			value+=500000;	/* if I have a change to win, I use it ;-) */
			continue;
		}  
		else if(other_dir.SameColor>=4)
		{
			value+=100000;	/* avoid losing where ever possible */
			continue;
		}

		v=getAreaValue(&me_dir);	 
		if(goGame->Player[playerNr].PlayMode==PLAY_OFFENSIVE)
		{
			v*=11;
			v/=10;
		}
		value+=v;
    
		v=getAreaValue(&other_dir);
		if(goGame->Player[playerNr].PlayMode==PLAY_DEFENSIVE)
		{
			v*=11;
			v/=10;
		}
		value+=v;
	}
  
	return(value);
}


/************************************************************************
 *																		*
 * get value of the area in the specified direction					 	*
 *																		*
 ************************************************************************/

static long getAreaValue(DirInfo *di)
{
	long	value=0;
	int	h;  

	if(di->OpenEnds==0 || di->SizeOfArea<5)
		return 0;
  
	value=1;
	h=di->SameColor;
	while(h--)
		value*=10;

	h=di->SameInArea;
	while(h--)
		value*=3;
  
	if(di->OpenEnds>1)
		value*=11;
 
	if(di->SizeOfArea>5)
		value*=2;
  
	return value;
}


/************************************************************************
 *																		*
 * calculate the next move for player playerNr 						 	*
 *																		*
 ************************************************************************/

static int dirAdd[][2]=
{
	{ 1, 0 },
	{ 1, -1 },
	{ 0, 1 },
	{ 1, 1 }
};

static void getDirectionInfo(int dir, int x, int y, int playerNr,
			     DirInfo *di)
{
	int 	x_add, y_add, x_pos, y_pos;
	int 	i, dirs_done;
  
	x_add=dirAdd[dir][0];
	y_add=dirAdd[dir][1];
  
	di->SameColor=0;
	di->SameInArea=0;
	di->SizeOfArea=1;
	di->OpenEnds=0;
	dirs_done=0; 
  
	while(dirs_done<2)
	{  
		i=0;
		x_pos=x+x_add;
		y_pos=y+y_add;
		while(getField(x_pos, y_pos)==(playerNr+1))
		{
			i++;
			di->SizeOfArea++;
			di->SameColor++;
    
			x_pos+=x_add;
			y_pos+=y_add;
		}
  
		if(getField(x_pos, y_pos)==FIELD_EMPTY)
		{
			i++;
			di->OpenEnds++;
			di->SizeOfArea++;

			x_pos+=x_add;
			y_pos+=y_add;

			while(5-i>0 && getField(x_pos, y_pos)==(playerNr+1))
			{
				i++;
				di->SizeOfArea++;
				di->SameInArea++;

				x_pos+=x_add;
				y_pos+=y_add;
			}
      
			while(5-i>0 && (getField(x_pos, y_pos)==FIELD_EMPTY))
			{
				i++;
				di->SizeOfArea++;
	
				x_pos+=x_add;
				y_pos+=y_add;
			}
		}
    
		x_add=-x_add;
		y_add=-y_add;
		dirs_done++;
	}  
}


/************************************************************************
 *																		*
 * check if player has won by making move <x>, <y> 					 	*
 *																		*
 ************************************************************************/

static gboolean checkFor5(int x, int y, int *x1, int *y1, int *x2, int *y2)
{
	DirInfo	di;
	int	k, player;
 
	if(getField(x,y)==FIELD_PLAYER1)
		player=0;
	else if(getField(x,y)==FIELD_PLAYER2)
		player=1;
	else
		return FALSE;
  
	for(k=0;k<4;k++)
	{
		getDirectionInfo(k, x, y, player, &di);
			     
		if(di.SameColor>=4)
		{ 
			*x1=x;
			*y1=y;
			while(getField(*x1+dirAdd[k][0], *y1+dirAdd[k][1])==player+1)
			{
				*x1+=dirAdd[k][0];
				*y1+=dirAdd[k][1];
			}
      
			*x2=x;
			*y2=y;
			while(getField(*x2-dirAdd[k][0], *y2-dirAdd[k][1])==player+1)
			{
				*x2-=dirAdd[k][0];
				*y2-=dirAdd[k][1];
			}
            
			return TRUE;
		}  
	}
  
	return FALSE;
}
