/* GnoBang: (c) 2008 by Andreas S. Krebs
 * 
 * This file is part of GnoBang.
 *
 * GnoBang is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * GnoBang is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GnoBang.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "gnobang.h"
    
/* local prototypes */

static void setMenuItemsSensitive(gboolean sens);
static gboolean timerCB(gpointer data);

/************************************************************************
 *																		*
 * Callback procedure for the `new`-menu item:  						*
 *   this procedure opens a message box and asks for a confirmation		*
 *   to start a new game												*
 *																		*
 ************************************************************************/

void newQuestionCB(GtkObject *object)
{
	if(goGame->GameStatus==GAME_FINISHED || goGame->GameStatus==GAME_NEW)
	{
		newGame();
	}
	else
	{
		GtkWidget	*dialog;
		gint		rc;
		
		if(goGame->GameStatus==GAME_RUNNING)
			setGameStatus(GAME_PAUSED);
		
		dialog=gtk_message_dialog_new_with_markup(gnoBang->window,
	            GTK_DIALOG_DESTROY_WITH_PARENT,
	            GTK_MESSAGE_QUESTION,
	            GTK_BUTTONS_YES_NO,
	            "<span size=\"large\">Do you want to start a new game?</span>");
		
		gtk_window_set_title(GTK_WINDOW(dialog), "New Game?");
		rc=gtk_dialog_run(GTK_DIALOG(dialog));
		gtk_widget_destroy(dialog);
		
		if(rc==GTK_RESPONSE_YES)
		{
			if(goGame->GameStatus==GAME_RUNNING || goGame->GameStatus==GAME_PAUSED)
				playSound(NEW_SOUND, FALSE);
		 
			newGame();
		}
	}
}


/************************************************************************
 *                                                                      *
 * set widget sensitive True, to enable widget                          *
 *                                                                      *
 ************************************************************************/

void switchOnWidget(int widgetID)
{
	GtkWidget *widget;  

  	if(goGame->CheatMode==CHEAT_NONE || (widgetID!=HINT_WIDGET && goGame->CheatMode!=CHEAT_UNDO))
  		return;
    
	if(widgetID==UNDO_WIDGET)
		widget=GTK_WIDGET(gnoBang->undoButton);
	else if(widgetID==REDO_WIDGET)
		widget=GTK_WIDGET(gnoBang->redoButton);
	else if(widgetID==HINT_WIDGET)
		widget=GTK_WIDGET(gnoBang->hintButton);
	else
		return;

	gtk_widget_set_sensitive(widget, TRUE);
}


/************************************************************************
 *                                                                      *
 * set widget sensitive False, to disable widget                        *
 *                                                                      *
 ************************************************************************/

void switchOffWidget(int widgetID)
{
	GtkWidget *widget;  
  
	if(widgetID==UNDO_WIDGET)
		widget=GTK_WIDGET(gnoBang->undoButton);
	else if(widgetID==REDO_WIDGET)
		widget=GTK_WIDGET(gnoBang->redoButton);
	else if(widgetID==HINT_WIDGET)
		widget=GTK_WIDGET(gnoBang->hintButton);
	else
		return;

	gtk_widget_set_sensitive(widget, FALSE);
}


/************************************************************************
 *                                                                      *
 * set menu item sensitive, to disable or enable sub menus              *
 *                                                                      *
 ************************************************************************/

static void setMenuItemsSensitive(gboolean sens)
{
	gtk_action_set_sensitive(gnoBang->prefsMenuAction, sens);
}


/************************************************************************
 *																		*
 * change label of the `pause`-button depending on the current 			*
 * GameStatus															*
 *																		*
 ************************************************************************/

static char *statusText[]= 
{ 
	"Start",
	"Pause",
	"Continue",
	"New"
};

static char *stockImage[]= 
{ 
	GTK_STOCK_MEDIA_PLAY,
	GTK_STOCK_MEDIA_PAUSE,
	GTK_STOCK_MEDIA_NEXT,
	GTK_STOCK_MEDIA_STOP
};

void setGameStatus(int status)
{
	goGame->GameStatus=status;  

	if(status==GAME_RUNNING)
	{
		switchOnWidget(HINT_WIDGET);
		switchOffWidget(REDO_WIDGET);
	}
	else
		switchOffWidget(HINT_WIDGET);
	  
	if(status==GAME_NEW)
		setMenuItemsSensitive(TRUE);
	else
		setMenuItemsSensitive(FALSE);
	
	if(status==GAME_FINISHED || status==GAME_PAUSED)
	{
		setPlayerTimer(0);	/* redisplay both timers */
		setPlayerTimer(1);

		setPlayersTurn(goGame->PlayersTurn);	/* redisplay players turn */  
	}

	gtk_label_set_text(gnoBang->controlLabel, statusText[status]);
	gtk_image_set_from_stock(gnoBang->controlImage, stockImage[status], GTK_ICON_SIZE_BUTTON);
}


/************************************************************************
 *																		*
 * set the players turn, by changing the label of the tomove widget 	*
 *																		*
 ************************************************************************/

void setPlayersTurn(int actPlayer)
{
	gchar	*txt=g_markup_printf_escaped("<span size=\"large\" color=\"#%s\">%s</span>", gnoBang->playerColor[actPlayer], goGame->Player[actPlayer].Name);
	gtk_label_set_markup(gnoBang->toMoveLabel, txt);
	g_free(txt);
		
	setCursor(actPlayer);
	goGame->PlayersTurn=actPlayer;  
	goGame->Player[actPlayer].TimeTurnStart=time(NULL);
	
	/* now start timer if not yet started */
	startPlayerTimer();
}


/************************************************************************
 *																		*
 * start game timer for player <playerNr> 								*
 *																		*
 ************************************************************************/

static guint timerID=0;

void startPlayerTimer()
{
	if(timerID==0)	/* only start if not already running */
	{
		/* initial set both turn times to invalid */
		goGame->Player[0].TimeTurnStart=(time_t)-1;
		goGame->Player[1].TimeTurnStart=(time_t)-1;
		
		timerID=gdk_threads_add_timeout(500, timerCB, NULL);
	}
}


/************************************************************************
 *																		*
 * Callback procedure for timer:  										*
 *   this procedure updates the players clock and starts the timer		*
 *   again if necessary													*
 *																		*
 ************************************************************************/

static gboolean timerCB(gpointer data)
{
	if(goGame->GameStatus==GAME_RUNNING) 
	{
		int	playerNr=goGame->PlayersTurn;
		time_t	now=time(NULL);
		time_t	elapsed;
		
		if(goGame->Player[playerNr].TimeTurnStart==-1)
			elapsed=0;
		else
			elapsed=now-goGame->Player[playerNr].TimeTurnStart;
		
		goGame->Player[playerNr].Seconds+=elapsed;
		goGame->Player[playerNr].TimeTurnStart=now;

		if(elapsed>0)
			setPlayerTimer(playerNr);

		return TRUE;		// call again
	}
	else   
	{
		timerID=0;
		return FALSE;		// destroy timer
	}
}


/************************************************************************
 *																		*
 * change label of clock widget, to display new time 					*
 *																		*
 ************************************************************************/

void setPlayerTimer(int playerNr)
{
	static char timetxt[10]="00:00:00\0";
	
	int		n;
	long	secs;
	
	secs=goGame->Player[playerNr].Seconds;

	n=secs/3600;	/* number of hours */
	secs%=3600;
	timetxt[0]=n/10+'0';
	timetxt[1]=n%10+'0';
 
	n=secs/60;
	timetxt[3]=n/10+'0';
	timetxt[4]=n%10+'0';
 
	n=secs%60;
	timetxt[6]=n/10+'0';
	timetxt[7]=n%10+'0';
	
	gchar	*txt=g_markup_printf_escaped("<span size=\"large\" color=\"#%s\">%s</span>", gnoBang->playerColor[playerNr], timetxt);
	gtk_label_set_markup(gnoBang->playerTimer[playerNr], txt);
	g_free(txt);
}
